#include "grammar.h"

#include <stdlib.h>

// Returns an index into an array from a 3D index
static unsigned int linearIndex(const glm::ivec3& i, const glm::ivec3& d)
{
    return (d.x * d.y * i.z) + (d.x * i.y) + i.x;
}

unsigned int Grammar::addSymbol(const Symbol& symbol)
{
    symbols.push_back(symbol);
    return symbols.size() - 1;
}

unsigned int Grammar::addTransform(const Transform& transform)
{
    transforms.push_back(transform);
    return transforms.size() - 1;
}

void Grammar::addRule(unsigned int symbolIndex, unsigned int transformIndex)
{
    ruleLookup[symbolIndex].push_back(transformIndex);
}

void Grammar::parse(std::vector<float>& voxelGrid, const glm::ivec3& gridDim, unsigned int iterations, const glm::ivec3& stride)
{
    for(unsigned int i = 0; i < symbols.size(); ++i)
    {
        if(ruleLookup.find(i) == ruleLookup.end())
        {
            continue;
        }

        const Symbol& symbol = symbols[i];
        const std::vector<unsigned int>& transformIndices = ruleLookup[i];

        for(int z = 0; z < gridDim.z; ++z)
        {
            for(int y = 0; y < gridDim.y; ++y)
            {
                for(int x = 0; x < gridDim.x; ++x)
                {
                    glm::ivec3 voxelIndex(x, y, z);
                    if(checkSymbolMatch(symbol, voxelGrid, voxelIndex, gridDim))
                    {
                        unsigned int tidx = transformIndices[rand() % transformIndices.size()];
                        transform(transforms[tidx], voxelGrid, voxelIndex, gridDim);
                    }
                }
            }
        }
    }
}

bool Grammar::checkSymbolMatch(const Symbol& s, std::vector<float>& voxelGrid, const glm::ivec3& voxelGridIndex, const glm::ivec3& voxelGridDim)
{
    bool result = false;
    for (int z = 0; z < s.dim.z; ++z)
    {
        for (int y = 0; y < s.dim.y; ++y)
        {
            for (int x = 0; x < s.dim.x; ++x)
            {
                glm::ivec3 sidx(x, y, z);
                const auto& op = s.ops[linearIndex(sidx, s.dim)];
                float voxelValue = voxelGrid[linearIndex(voxelGridIndex + sidx, voxelGridDim)];
                switch (op.first)
                {
                case Symbol::IGN:
                    result = result && true;
                    break;

                case Symbol::OCCUPIED:
                    result = result && (voxelValue > 0.0f);
                    break;

                case Symbol::UNOCCUPIED:
                    result = result && (voxelValue <= 0.0f);
                    break;

                case Symbol::EQ:
                    result = result && (voxelValue == op.second);
                    break;

                case Symbol::NEQ:
                    result = result && (voxelValue != op.second);
                    break;

                case Symbol::LT:
                    result = result && (voxelValue < op.second);
                    break;

                case Symbol::LEQ:
                    result = result && (voxelValue <= op.second);
                    break;

                case Symbol::GT:
                    result = result && (voxelValue > op.second);
                    break;

                case Symbol::GEQ:
                    result = result && (voxelValue >= op.second);
                    break;

                default:
                    break;
                }

                if (!result)
                {
                    return false;
                }
            }
        }
    }

    return true;
}

void Grammar::transform(const Transform& t, std::vector<float>& voxelGrid, const glm::ivec3& voxelGridIndex, const glm::ivec3& voxelGridDim)
{
    for (int z = 0; z < t.dim.z; ++z)
    {
        for (int y = 0; y < t.dim.y; ++y)
        {
            for (int x = 0; x < t.dim.x; ++x)
            {
                glm::ivec3 tidx(x, y, z);
                const auto& op = t.ops[linearIndex(tidx, t.dim)];
                switch (op.first)
                {
                case Transform::NOP:
                    break;

                case Transform::SET:
                    voxelGrid[linearIndex(voxelGridIndex + tidx, voxelGridDim)] = op.second;
                    break;

                case Transform::ADD:
                    voxelGrid[linearIndex(voxelGridIndex + tidx, voxelGridDim)] += op.second;
                    break;

                default:
                    break;
                }
            }
        }
    }
}